"""Loggers

Loggers module.
"""

import logging
import sys
import time
import weakref
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from logging import Logger as BaseLogger
from logging import _lock
from logging.handlers import RotatingFileHandler
from os import getpid, path
from typing import List, Literal, Optional, Union

import ecs_logging
from graypy import GELFHTTPHandler, GELFTCPHandler, GELFUDPHandler
from graypy.handler import BaseGELFHandler
from pythonjsonlogger import jsonlogger
from typing_extensions import final

try:
    import httplib
except ImportError:
    import http.client as httplib

DEBUG_DEFAULT_FORMAT_STRING = (
    "[{process:07d} {asctime}] {levelname} [{filename} {lineno}]: {appName} {requestId}{channel}: {message}"
)

# (str): format string for non debug logs
NOT_DEBUG_DEFAULT_FORMAT_STRING = "[{process:07d} {asctime}] {levelname}: {appName} {requestId}{channel}: {message}"

DEBUG_DEFAULT_JSON_FIELDS = (
    "{process} {asctime} {levelname} {filename} {lineno} {appName} {requestId} {channel} {message}"
)

NOT_DEBUG_DEFAULT_JSON_FIELDS = "{process} {asctime} {levelname} {appName} {requestId} {channel} {message}"


# (int): size of one megabyte in bytes
SIZE_OF_MB = 2**20
TIME_ZONE_DELTA = time.timezone  #: correction for getting timestamp in utc


class GELFHTTPSHandler(GELFHTTPHandler):
    """Gelf https log implementation"""

    def emit(self, record: logging.LogRecord):
        """Convert log record to GELF and emit it to Graylog
        via a HTTP POST request

        Args:
            record: log record to convert
        """
        pickle = self.makePickle(record)
        connection = httplib.HTTPSConnection(host=self.host, port=self.port, timeout=self.timeout)
        connection.request("POST", self.path, pickle, self.headers)


GELF_HANDLERS_MAP = {
    "tcp": GELFTCPHandler,
    "udp": GELFUDPHandler,
    "http": GELFHTTPHandler,
    "https": GELFHTTPSHandler,
}


@dataclass
class GelfSettings:
    """Gelf logging settings"""

    # gelf input protocol
    type: str = "udp"
    # gelf input host
    host: str = "127.0.0.1"
    # gelf input port
    port: str = "9000"


def _adaptChannelForRecordFormat(channel: str) -> str:
    """Adopt channel name for record"""
    return f": {channel}" if channel else ""


def generateStdoutLogHandler() -> logging.StreamHandler:
    """
    Generate stout log handler
    Returns:
        new stdout handler
    """
    handler = logging.StreamHandler(sys.stdout)
    return handler


def _getFormatter(
    formatter: Union[Literal["default", "json"]], level: int, multilineStackTrace: bool
) -> logging.Formatter:
    """
    Get log formatter
    Args:
         formatter: formatter type or instance of logging.Formatter
         level: logging level
         multilineStackTrace: allow a multiline line stack trace message
    """
    if formatter == "default":
        if level == logging.DEBUG:
            fmtString = DEBUG_DEFAULT_FORMAT_STRING
        else:
            fmtString = NOT_DEBUG_DEFAULT_FORMAT_STRING
        formatter_cls = Formatter if multilineStackTrace else OneLineExceptionFormatter
        formatter = formatter_cls(fmt=fmtString, style="{")
    elif formatter == "json":
        if level == logging.DEBUG:
            fmtString = DEBUG_DEFAULT_JSON_FIELDS
        else:
            fmtString = NOT_DEBUG_DEFAULT_JSON_FIELDS
        formatter = JsonFormatter(fmtString, style="{")
    elif formatter == "ecs":
        formatter = ECSFormatter(style="{")
    else:
        raise ValueError(f"Unknown formatter type: {formatter}")
    return formatter


def setUpHandler(
    handler: logging.Handler,
    multilineStackTrace: bool,
    level: int,
    formatter: Union[Literal["default", "json"], logging.Formatter],
):
    """
    Setup log handler. set log record formatter, level
    Args:
        handler: handler
        multilineStackTrace: allow a multiline line stack trace message
        level: log level
        formatter: log formatter
    """
    handler.level = level
    formatter = _getFormatter(formatter, level, multilineStackTrace) if isinstance(formatter, str) else formatter
    handler.setFormatter(formatter)


def generateFileHandler(logFileName: str, maxSize: int = 1024) -> logging.FileHandler:
    """
    Generate file handler
    Args:
        logFileName: filename
        maxSize: max size of log file

    Returns:
        rotating file  handler if maxSize> 0 otherwise usual file handler
    """
    if maxSize > 0:
        handler = RotatingFileHandler(logFileName, maxBytes=maxSize * SIZE_OF_MB, backupCount=5)
    else:
        handler = logging.FileHandler(logFileName)
    return handler


def generateGelfHandler(gelfSettings: GelfSettings) -> BaseGELFHandler:
    """
    Generate gelf handler
    Args:
        gelfSettings: gelf settings

    Returns:
        gelf handler

    Raises:
        ValueError: If unknown type received
    """
    if gelfSettings.type not in GELF_HANDLERS_MAP:
        raise ValueError(f"Unknown gelf log type {gelfSettings.type}")
    return GELF_HANDLERS_MAP[gelfSettings.type](host=gelfSettings.host, port=gelfSettings.port)


def getLogTimestamp(timeIsLocal: bool) -> int:
    """
    Generate timestamp for logs

    Args:
        timeIsLocal: timestamp generate in localtime or not
    Returns:
        timestamp with
    """
    if timeIsLocal:
        timestamp = int(time.time() - TIME_ZONE_DELTA)
    else:
        timestamp = int(time.time())
    return timestamp


def _getDefaultRequestId() -> str:
    """
    Get default request id for record which is initiated of the service.

    Format '{ts_start_service},00000000-0000-4000-a000-{pid}'

    Returns:
        request id
    """
    timestamp = getLogTimestamp(timeIsLocal=True)
    return f"{timestamp:08d},00000000-0000-4000-a000-{getpid():012d}"


DEFAULT_REQUEST_ID = _getDefaultRequestId()  #: request id for record which is initiated of the service
requestIdCtx = ContextVar("requestId", default=None)


class Formatter(logging.Formatter):
    """
    Log record formater
    """

    # time converter
    converter = time.localtime
    # default milliseconds format
    default_msec_format = "%s.%03d000"


class JsonFormatter(Formatter, jsonlogger.JsonFormatter):
    """
    Json log record formatter
    """

    def process_log_record(self, log_record: dict):
        """Postprocess json log record fields"""
        if log_record.get("channel") == "":
            log_record.pop("channel")
        return log_record


class ECSFormatter(Formatter, ecs_logging.StdlibFormatter):
    """
    ECS log record formatter
    """

    def format(self, record):
        """Preprocess log records before ECS formatting"""
        if hasattr(record, "requestId"):
            setattr(record, "http.request.id", record.requestId)
            del record.requestId
        if hasattr(record, "channel"):
            del record.channel
        if hasattr(record, "appName"):
            setattr(record, "service.name", record.appName)
            del record.appName
        return super().format(record)


class OneLineExceptionFormatter(Formatter):
    """
    One line exception log record formatter

    References:
        https://docs.python.org/3/howto/logging-cookbook.html#customized-exception-formatting
    """

    def format(self, record: logging.LogRecord) -> str:
        """
        Format an exception so that it prints on a single line.
        """
        msg = super().format(record)
        if record.exc_text:
            msg = msg.replace("\n", "    |")
        return msg


@dataclass
class LogSettings:
    """
    Logger settings container
    """

    # app name
    appName: str
    # log level
    logLevel: int
    # log records time
    logTime: str
    # folder for logs
    folderForLog: str
    # max size of log file
    maxSize: int
    # log file suffix
    logNameSuffix: Optional[str]
    # write or not logs to stdout
    logToStdOut: bool
    # write or not logs to file
    logToFile: bool
    # write or not logs to gelf
    logToGelf: bool
    # gelf settings:
    gelfSettings: GelfSettings
    # allow a multiline line stack trace message
    multilineStackTrace: bool
    # preferred log formatting type
    formatter: Union[Literal["default", "json", "ecs"], logging.Formatter]


class LogDest(Enum):
    """
    Log destination
    """

    file = "file"  #: destination is log file
    stdout = "stdout"  #: destination is stdout


class Logger(BaseLogger):
    """
    Application logger.

    Global logger class. This class realize following features:

    1. Class contains global handlers. All loggers will be to have this handlers.

    2. Class replaces itself default logger class from standard library (logging.Logger). All third party loggers
    (based on logging.Logger) will be use handlers of this class

    3. method initiate updates all created loggers.

    All features based on class variable `_handlers` and property `handlers`. They patch attribute `handlers` of
    logger from standard library and make them globals.


    >>> from logging import getLogger
    >>> import log
    >>>
    >>> lunaSdkLogger = getLogger("luna.sdk")
    >>> thirdPartyLogger = getLogger("thirdpartylogger")
    >>> strictLogger = getLogger("platform")
    >>> subStrictLogger = getLogger("platform.api")

    >>> log.initiate(
    ...     appName="test",
    ...     logLevel="INFO",
    ...     logTime="LOCAL",
    ...     folderForLog="./",
    ...     maxSize=1024,
    ...     logToStdOut=True,
    ...     logToFile=False,
    ...     multilineStackTrace=True,
    ...     thirdPartyLoggerWhiteList=["luna.*", "platform"],
    ... )
    >>> lunaSdkLogger.info("sdk logger")  # this message will be logged
    [0025084 2022-06-21 12:09:59.000379] INFO: test 1655813399,00000000-0000-4000-a000-000000025084:
    luna.sdk: sdk logger
    >>> thirdPartyLogger.info("third party logger")  # this message will be skipped
    >>> thirdPartyLogger.warning("warning third party logger")  # this message will be logged
    [0025084 2022-06-21 12:09:59.000381] WARNING: test 1655813399,00000000-0000-4000-a000-000000025084:
    thirdpartylogger: warning third party logger
    >>> strictLogger.info("strict logger")  # this message will be logged
    [0026856 2022-06-21 12:12:18.000805] INFO: test 1655813538,00000000-0000-4000-a000-000000026856: platform:
    strict logger
    >>> subStrictLogger.info("warning third party logger")  # this message will be skipped
    """

    # logger settings
    _settings: Optional[LogSettings] = None
    # logs fie handlers
    _logsFileHandlers: List[logging.FileHandler] = []
    # stdout handler
    _stdoutHandler: Optional[logging.StreamHandler] = None
    # gelf handler
    _gelfHandler: Optional[BaseGELFHandler] = None
    # logger default request id
    defaultRequestId: str = DEFAULT_REQUEST_ID
    # global logger handlers
    _handlers = []

    # third party logger whit list. allowed format:
    #               - strict name
    #               - wildcard `rootname.*`
    _thirdPartyLoggerWhiteList: list[str] = []

    @classmethod
    def checkThirdPartyLoggerName(cls, name) -> bool:
        """
        Check third party logger name that is allowed.

        Args:
            name:  third party logger name
              - strict name
              - wildcard `rootname.*`

        Returns:
            true if name matches with any name from thirdPartyLoggerWhiteList otherwise false
        """
        for allowedLoggerName in cls._thirdPartyLoggerWhiteList:
            if allowedLoggerName.endswith(".*"):
                rootLoggerName = allowedLoggerName[:-2]
                if name.startswith(f"{rootLoggerName}.") or name == rootLoggerName:
                    return True
            elif allowedLoggerName == name:
                return True
        return False

    @staticmethod
    def getLogLevel(logLevel) -> int:
        """
        Get log level from config for logger.

        Returns:
            int: if logLevel not set or incorrect will return logging.NOTSET
        """
        if logLevel == "DEBUG":
            return logging.DEBUG
        if logLevel == "ERROR":
            return logging.ERROR
        if logLevel == "INFO":
            return logging.INFO
        if logLevel == "WARNING":
            return logging.WARNING
        return logging.NOTSET

    @classmethod
    @final
    def addHandler(cls, hdlr: logging.Handler) -> None:
        """
        Add handlers to all loggers

        Overwrite original method. Before it was instance method, now - global classmethod
        Args:
            hdlr: log handler
        """
        with _lock:
            if hdlr not in cls._handlers:
                cls._handlers.append(hdlr)

    @classmethod
    @final
    def removeHandler(cls, hdlr: logging.Handler):
        """
        Remove the specified handler from all loggers.
        Overwrite original method. Before it was instance method, now - global classmethod

        Args:
            hdlr: log handler
        """
        with _lock:
            if hdlr in cls._handlers:
                cls._handlers.remove(hdlr)

    @property
    def logTime(self) -> str:
        """
        Get current log time format
        Returns:
            logTime from settings
        """
        return self._settings.logTime

    @property
    def level(self) -> int:
        """
        Overwrite property. Not save log level to an in instance of the Logger for correct reinitialize loggers.

        Returns:
            class log level
        """
        if not self._settings:
            initiate(logLevel="DEBUG", logToStdOut=True, logToFile=False)
            self.warning("Logger used has not been initialized. DEBUG level is set by default.")
        return self._settings.logLevel

    @level.setter
    def level(self, level: int):
        """
        Set log level stub. need for supper().__init__
        Args:
            level: new log level

        Warnings:
            Use only `initiate` and `reInitialize` functions
        """

    @property
    def requestId(self) -> str:
        """
        Get current request id for logger instance

        Returns:
            request id. priority: 1) request id from __init__ 2) default
        """
        return self._requestId or requestIdCtx.get() or self.defaultRequestId

    @classmethod
    def getClassRequestId(cls) -> str:
        """
        Get request id on class. Usually used for third party logger (__init__ has not any request id parameter for
        these loggers)
        """
        return requestIdCtx.get() or cls.defaultRequestId

    @classmethod
    def _defaultFilter(cls, record: logging.LogRecord) -> bool:
        """
        Add extra data (request id and app name) to record. Usually used for third party logger. These loggers
        do not contain client defined parameters `requestId` and `channel`. This filter must be set up to every handler.

        Args:
            record: record

        Returns:
            True
        """
        record.appName = cls._settings.appName
        if not hasattr(record, "requestId"):
            record.requestId = cls.getClassRequestId()
        if not hasattr(record, "channel"):
            if cls._settings.formatter == "default":
                record.channel = _adaptChannelForRecordFormat(record.name)
            else:
                record.channel = record.name
        return True

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Add extra data (request id and app name) to record
        Args:
            record: record

        Returns:
            result of filters
        """
        record.appName = self._settings.appName
        record.requestId = self.requestId
        record.channel = (
            _adaptChannelForRecordFormat(self.channel) if self._settings.formatter == "default" else self.channel
        )
        return super().filter(record)

    @property
    @final
    def handlers(self) -> List[logging.Handler]:
        """
        Get list of global handlers. Rewrite local handlers from logging.Logger
        Not save handlers to an in instance of the Logger for correct reinitialize loggers. All created loggers
        have these handlers.

        Returns:
            list of global handlers.
        """
        return self._handlers

    @handlers.setter
    @final
    def handlers(self, handlers: object):
        """
        Set handlers stub. need for supper().__init__
        Args:
            handlers: new handlers

        Warnings:
            Use only `initiate` and `reInitialize` functions
        """

    def exception(self, *args, **kwargs) -> None:  #: pylint: disable-msg=W0221
        """
        Exception without msg support.
        Args:
            *args: args
            **kwargs: kwargs
        """
        if len(args) == 0 and "msg" not in kwargs:
            return super().exception("Uncaught exception occurred", *args, **kwargs)
        return super().exception(*args, **kwargs)

    instances = {}

    def __init__(self, template: str = "", requestId: Optional[str] = None) -> None:
        # pylint: disable-msg=W0613
        """
        Init logger.

        Args:
            requestId: user defined request id
            template: string for marking logs. Typical usage - request id.
        """
        self._requestId = requestId
        self.channel = template
        super().__init__(template)
        self.instances[id(self)] = weakref.ref(self)

    def __del__(self):
        self.instances.pop(id(self))


def initiate(
    appName: str = "",
    logLevel: str = "DEBUG",
    logTime: str = "LOCAL",
    folderForLog: str = "./",
    maxSize: int = 1024,
    logNameSuffix: Optional[str] = None,
    logToStdOut: bool = True,
    logToFile: bool = False,
    logToGelf: bool = False,
    gelfSettings: Optional[dict] = None,
    multilineStackTrace: bool = True,
    thirdPartyLoggerWhiteList: Optional[list[str]] = None,
    formatter: Union[Literal["default", "json", "ecs"], logging.Formatter] = "default",
) -> None:
    """
    (Re)initiate logger settings.

    Args:
        appName: application name
        logLevel: log level
        logTime: time of logs
        folderForLog: folder with log-files
        maxSize: max log file size, 1024 mb for default, 0 - disable rotate
        logNameSuffix: optional suffix name for log name files.
        logToStdOut: flush log to stdout
        logToGelf: flush log to a gelf input
        gelfSettings: gelf logging settings
        logToFile: flush log to a log file
        multilineStackTrace: allow a multiline line stack trace message
        thirdPartyLoggerWhiteList: third party logger whit list. allowed format:
           - strict name
           - wildcard `rootname.*`
        formatter: log formatter
    """

    def update3partyLogger(thirdPartyLogger: logging.Logger):
        """
        Update 3-party logger. set up handlers, log level and other
        Args:
            thirdPartyLogger: 3-party logger
        """
        if not Logger.checkThirdPartyLoggerName(thirdPartyLogger.name):
            thirdPartyLogger.setLevel(max(logging.WARNING, Logger._settings.logLevel))
        else:
            thirdPartyLogger.setLevel(Logger._settings.logLevel)
        thirdPartyLogger.propagate = False

        # close old handlers
        for handler in thirdPartyLogger.handlers:
            handler.close()
        # setup global hendlers
        with _lock:
            thirdPartyLogger.handlers = Logger._handlers

    def updateDefaultPythonLoggers():
        """
        setup this  class as default class for 3-party loggers, which use logging.getLogger, manager.getLogger,
        getLoggerClass.

        """
        # upgrade already created loggers
        for thirdPartyLogger in logging.Logger.manager.loggerDict.values():
            if not isinstance(thirdPartyLogger, BaseLogger):
                continue
            update3partyLogger(thirdPartyLogger)

        # set up class for future loggers
        loggerClass = logging.getLoggerClass()

        class ThirdPartyLoggers(loggerClass):
            """Third party logger factory"""

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.propagate = False
                if Logger.checkThirdPartyLoggerName(self.name):
                    self.level = Logger._settings.logLevel
                else:
                    self.level = max(logging.WARNING, Logger._settings.logLevel)
                with _lock:
                    self.handlers = Logger._handlers

        for logger in Logger.instances.values():
            logger()._cache.clear()

        Logger.manager.setLoggerClass(ThirdPartyLoggers)
        logging.setLoggerClass(ThirdPartyLoggers)

    def reInitializeStdoutLogging(settings: LogSettings):
        """
        Re-Initialize stdout log handlers. Closes old handlers and create new if it iis needed
        Args:
            settings: settings for re-initialization
        """
        if Logger._stdoutHandler:
            Logger.removeHandler(Logger._stdoutHandler)
            if not Logger._stdoutHandler.stream.closed:
                # Let capsys monkeypatch the stream
                Logger._stdoutHandler.close()
            Logger._stdoutHandler = None

        if settings.logToStdOut:
            Logger._stdoutHandler = generateStdoutLogHandler()
            setUpHandler(
                Logger._stdoutHandler,
                multilineStackTrace=settings.multilineStackTrace,
                level=settings.logLevel,
                formatter=settings.formatter,
            )
            Logger.addHandler(Logger._stdoutHandler)

    def reInitializeGelfLogging(settings: LogSettings):
        """
        Re-Initialize gelf logs handlers. Closes old handlers and create new if it iis needed
        Args:
            settings: settings for re-initialization
        """
        if Logger._gelfHandler:
            Logger._gelfHandler.close()
            Logger.removeHandler(Logger._gelfHandler)
            Logger._gelfHandler = None

        if settings.logToGelf:
            Logger._gelfHandler = generateGelfHandler(settings.gelfSettings)
            # Logger._gelfHandler.level = settings.logLevel
            setUpHandler(
                Logger._gelfHandler,
                multilineStackTrace=settings.multilineStackTrace,
                level=settings.logLevel,
                formatter=settings.formatter,
            )
            Logger.addHandler(Logger._gelfHandler)

    def reInitializeFileLogging(settings: LogSettings):
        """
        Re-Initialize file logs handlers. Closes old handlers and create new if it iis needed
        Args:
            settings: settings for re-initialization
        """

        for handler in Logger._logsFileHandlers:
            handler.close()
            Logger.removeHandler(handler)

        Logger._logsFileHandlers = []
        if settings.logToFile:
            folderForLog = settings.folderForLog
            appName = settings.appName
            logNameSuffix = settings.logNameSuffix
            multilineStackTrace = settings.multilineStackTrace
            for fileLogLevel in ("INFO", "ERROR", "DEBUG", "WARNING"):
                if settings.logLevel > getattr(logging, fileLogLevel):
                    continue

                if not logNameSuffix:
                    logFileName = path.join(folderForLog, f"{appName}_{fileLogLevel}.txt")
                else:
                    logFileName = path.join(folderForLog, f"{appName}_{fileLogLevel}_{logNameSuffix}.txt")

                mStackTrace = multilineStackTrace if fileLogLevel != "DEBUG" else True
                handler = generateFileHandler(logFileName, settings.maxSize)
                setUpHandler(handler, mStackTrace, level=getattr(logging, fileLogLevel), formatter=settings.formatter)
                Logger.addHandler(handler)
                Logger._logsFileHandlers.append(handler)

    def reInitializeLogTime(settings: LogSettings):
        """
        Re Initialize log time. Update time converter of global formatter
        Args:
            settings: settings for re-initialization
        """
        Formatter.converter = time.localtime if settings.logTime == "LOCAL" else time.gmtime

    _settings = LogSettings(
        maxSize=maxSize,
        appName=appName,
        logTime=logTime,
        logLevel=Logger.getLogLevel(logLevel),
        logNameSuffix=logNameSuffix,
        logToFile=logToFile,
        logToStdOut=logToStdOut,
        logToGelf=logToGelf,
        gelfSettings=GelfSettings(**(gelfSettings or {})),
        folderForLog=folderForLog,
        multilineStackTrace=multilineStackTrace,
        formatter=formatter,
    )
    if _settings == Logger._settings and Logger._thirdPartyLoggerWhiteList == thirdPartyLoggerWhiteList:
        return
    Logger._thirdPartyLoggerWhiteList = thirdPartyLoggerWhiteList if thirdPartyLoggerWhiteList is not None else []
    try:
        reInitializeStdoutLogging(_settings)
        reInitializeFileLogging(_settings)
        reInitializeGelfLogging(_settings)
        reInitializeLogTime(_settings)
    except Exception:
        # fallback to old settings, try set correct Logger state for logging errors
        if Logger._settings is not None:
            reInitializeStdoutLogging(Logger._settings)
            reInitializeFileLogging(Logger._settings)
            reInitializeGelfLogging(Logger._settings)
            reInitializeLogTime(Logger._settings)
        raise
    Logger._settings = _settings
    for handler in Logger._handlers:
        handler.addFilter(Logger._defaultFilter)
    updateDefaultPythonLoggers()
